import java.util.Vector; 
import java.io.*; 

/* Copyright K Lano 2012-2018
   package: Utilities
*/ 

class Maplet
{ Object source;
  Object dest;  /* Usually both are named */

  Maplet(Object src, Object dst)
  { source = src;
    dest = dst; 
  }


  public Object getSource() 
  { return source; } 

  public Object getTarget()
  { return dest; } 

  public void print_maplet()
  { System.out.print(((Named) source).label);
    System.out.print("  |-->  ");
    System.out.println(((Named) dest).label); }

  public void print_maplet(PrintWriter out)
  { out.print(((Named) source).label);
    out.print("  |-->  ");
    out.println(((Named) dest).label); 
  }

  public String toString()
  { return source + "  |-->   " + dest; } 

  public void display() 
  { System.out.print(source); 
    System.out.print("  |-->  ");
    System.out.println(dest); }

  public boolean equals(Object mm) 
  { if (mm instanceof Maplet) 
    { Maplet maplet = (Maplet) mm; 
      return (source.equals(maplet.source) && 
              dest.equals(maplet.dest)); 
    } 
    else 
    { return false; } 
  } 
}

class Map
{ // really a multi-map

  int maxmap = 100;
  Vector elements;  /* vector of maplets */

  Map()  { elements = new Vector(maxmap); }

  Map(Vector elems)  { elements = elems; }

  public int size()
  { return elements.size(); } 

  public Maplet get(int i)
  { return (Maplet) elements.get(i); } 

  public void clear()
  { elements = new Vector(maxmap); }

  public void add_element(Maplet m)
  {  elements.addElement(m);  /* assumes still a map */
  }

  public void add_pair(Object o1, Object o2)
  { Maplet mm = new Maplet(o1,o2);
    if (elements.contains(mm)) { } 
    else 
    { elements.addElement(mm); }
  } 

  public void remove_pair(Object o1, Object o2)
  { Maplet mm = new Maplet(o1,o2);
    Vector mms = new Vector(); 
    mms.add(mm); 
    elements.removeAll(mms); 
  } 

  public static Map union(Map m1, Map m2)
  { Vector elems = m1.elements; 
    Vector elems2 = m2.elements; 
    Map res = new Map(); 

    for (int i = 0; i < elems.size(); i++) 
    { Maplet mm = (Maplet) elems.get(i);
      if (res.elements.contains(mm)) { } 
      else 
      { res.add_element(mm); }  
    }

    for (int i = 0; i < elems2.size(); i++) 
    { Maplet mm = (Maplet) elems2.get(i);
      if (res.elements.contains(mm)) { } 
      else 
      { res.add_element(mm); }  
    }

    return res; 
  } 

  public Map removeDuplicates() 
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++) 
    { Maplet mm = (Maplet) elements.get(i); 
      if (res.contains(mm)) { } 
      else 
      { res.add(mm); } 
    } 
    return new Map(res); 
  } 

 public void print_map()
 { for (int i = 0; i < elements.size(); i++)
   { ((Maplet) elements.elementAt(i)).print_maplet(); }
 }

 public String toString()
 { String res = ""; 

   for (int i = 0; i < elements.size(); i++)
   { res = res + " " + ((Maplet) elements.elementAt(i)) + "\n"; }
   return res; 
 }

  public boolean in_domain(Object x)
  { if (x == null) { return false; } 
    for (int i = 0; i < elements.size(); i++)
    { if (x.equals(((Maplet) elements.elementAt(i)).source))
      { return true; } 
    }
    return false; 
  }

  public boolean in_range(Object x)
  { if (x == null) { return false; } 
    for (int i = 0; i < elements.size(); i++)
    { if (x.equals(((Maplet) elements.elementAt(i)).dest))
      { return true; } 
    }
    return false; 
  }

  public Vector domain()
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++)
    { Maplet mm = (Maplet) elements.elementAt(i); 
      Object src = mm.source; 
      if (res.contains(src)) { } 
      else 
      { res.add(src); }
    } 
    return res; 
  }
 
  public Vector range()
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++)
    { Maplet mm = (Maplet) elements.elementAt(i); 
      Object src = mm.dest; 
      if (res.contains(src)) { } 
      else 
      { res.add(src); }
    } 
    return res; 
  }

  public Object apply(Object x)
  { if (x == null) { return null; } 

    for (int i = 0; i < elements.size(); i++)
    { if (x.equals(((Maplet) elements.elementAt(i)).source))
      { return ((Maplet) elements.elementAt(i)).dest; } 
    }
    return null; 
  }

  public static Vector compose(Vector map1, Vector map2)
  { Vector res = new Vector();
    for (int i = 0; i < map1.size(); i++)
    { Maplet mm1 = (Maplet) map1.elementAt(i);
      for (int j = 0; j < map2.size(); j++)
      { Maplet mm2 = (Maplet) map2.elementAt(j);
        if (mm1.dest.equals(mm2.source))
        { Maplet mm3 = new Maplet(mm1.source, mm2.dest);
          res.add(mm3); } } }
    return res; }

  public static Map compose(Map m1, Map m2)
  { Map res = new Map(); 
    res.elements = compose(m1.elements, m2.elements); 
    return res; 
  } 

  public void put(Object src, Object trg)
  { Map map2 = new Map(); 
    map2.add_pair(src,trg); 
    Vector newelems = compose(elements,map2.elements); 
    elements = VectorUtil.union(elements,newelems); 
  } 

  public static Vector transitiveClosure(Vector map)
  { Vector res = (Vector) map.clone();
    for (int i = 0; i < map.size(); i++)
    { Vector res1 = compose(res,map);
      Vector res2 = VectorUtil.vector_merge(res,res1);
      // System.out.println("Trans closure step " + i + ": " + res2);
      res = res2; 
    }
    return res; 
  }   

  public void addAll(Vector maplets)
  { elements.addAll(maplets); } 

  public void addAll(Map m2)
  { elements.addAll(m2.elements); } 

  public void removeAll(Vector maplets)
  { elements.removeAll(maplets); } 

  public static int countLoops(Map m)
  { Map m2 = new Map(); 
    int res = 0; 

    Map mnew = Map.compose(m,m); 
    while (mnew.size() > 0)
    { mnew.print_map(); 
      m2.clear(); 

      for (int i = 0; i < mnew.elements.size(); i++) 
      { Maplet mm = (Maplet) mnew.elements.get(i); 
        if (mm.source.equals(mm.dest)) 
        { System.out.println("Loop on " + mm.source); 
          res++; 
        }
        else 
        { m2.add_pair(mm.source, mm.dest); }
      }  
      mnew = Map.compose(m2,m); 
    } 
    return res; 
  } 
       
  public int countSelfMaps()
  { int res = 0; 
    for (int i = 0; i < elements.size(); i++) 
    { Maplet mm = (Maplet) elements.get(i); 
      if (mm.source != null && mm.source.equals(mm.dest)) 
      { res++; } 
    } 
    return res; 
  } 

  public Vector getSelfMaps()
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++) 
    { Maplet mm = (Maplet) elements.get(i); 
      if (mm.source != null && mm.source.equals(mm.dest)) 
      { if (res.contains(mm.source)) { } 
        else 
        { res.add(mm.source); } 
      } 
    } 
    return res; 
  } 


  public static Map domainRestriction(Vector dom, Map m)
  { Vector elems = new Vector(); 
    for (int i = 0; i < m.elements.size(); i++) 
    { Maplet mm = (Maplet) m.elements.get(i); 
      if (dom.contains(mm.source)) 
      { elems.add(mm); } 
    } 
    return new Map(elems); 
  } 

}

