import java.util.Vector; 

/* Copyright K. Lano 2015-2018
   Package: QVT
*/ 

public class Pattern
{ Vector bindsTo = new Vector(); // of Attribute
  Vector predicate = new Vector(); // of Expression

  public Object clone()
  { Pattern res = new Pattern(); 
    res.bindsTo = bindsTo; 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression e = (Expression) predicate.get(i); 
      Expression e1 = (Expression) e.clone(); 
      res.addPredicate(e1); 
    } 
    return res; 
  } 

  public void addBindsTo(Attribute b)
  { bindsTo.add(b); }

  public void addPredicate(Expression p)
  { if (p instanceof BinaryExpression) 
    { BinaryExpression be = (BinaryExpression) p; 
      if ("&".equals(be.operator))
      { addPredicate(be.left); 
        addPredicate(be.right); 
      }
      else 
      { predicate.add(be); } 
    } 
    else 
    { predicate.add(p); } 
  } 

  public boolean typeCheck(Vector types, Vector entities, Vector context, Vector env)
  { boolean res = true; 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression p = (Expression) predicate.get(i); 
      res = p.typeCheck(types,entities,context,env); 
    } 
    return res; 
  }  

  public String toString()
  { String res = ""; 
    for (int i = 0; i < predicate.size(); i++) 
    { res = res + predicate.get(i); 
      if (i < predicate.size() - 1)
      { res = res + " & "; } 
    } 
    return res; 
  }  

  public int complexity()
  { int res = 0; 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression p = (Expression) predicate.get(i); 
      res = res + p.syntacticComplexity(); 
    } 
    return res; 
  }  

  public int cyclomaticComplexity()
  { int res = 0; 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression p = (Expression) predicate.get(i); 
      res = res + p.cyclomaticComplexity(); 
    } 
    return res; 
  }  

  public Vector operationsUsedIn()
  { Vector res = new Vector(); 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression p = (Expression) predicate.get(i); 
      res.addAll(p.allOperationsUsedIn()); 
    } 
    return res; 
  }  

  public Vector variablesUsedIn()
  { Vector res = new Vector(); 
    for (int i = 0; i < predicate.size(); i++) 
    { Expression p = (Expression) predicate.get(i); 
      res.addAll(p.getVariableUses()); 
    } 
    return res; 
  }  

  public int epl() { return 0; } 
  // but should count let and ->iterate variables 


  public Expression toGuardCondition(Vector bound, Expression contextObj, Expression post)
  { return null; }

  public Expression toSourceExpression(Vector bound, Expression contextObj)
  { return null; }

  public Expression toTargetExpression(Vector bound, Expression contextObj)
  { return null; }

  public Expression toSourceExpressionOp(Vector bound, Expression contextObj)
  { return null; }

  public Expression toTargetExpressionOp(Vector bound, Expression contextObj)
  { return null; }


  public Vector whenToExp(Vector bound, Vector rules, Vector entities)
  { // R(s,t) is r$trace : s.traces$R$p1 & t = r$trace.p2
    // The s, t are bound at this point. 
    // An or of trace tests leads to multiple constraints, one for each disjunct. 

    Vector res = new Vector(); 

    Expression cond = new BasicExpression(true);
    res.add(cond); 

    for (int i = 0; i < predicate.size(); i++)
    { Expression conj = (Expression) predicate.get(i);
      Expression e = conj;
      if (conj != null && conj.isRuleCall(rules) && (conj instanceof BasicExpression))
      { BasicExpression be = (BasicExpression) conj; 
        e = be.testTraceRelationWhen(entities,bound); 

        // System.out.println("RULE TEST: " + e); 

        Vector pars = be.getParameters(); 
        for (int p = 0; p < pars.size(); p++) 
        { Expression par = (Expression) pars.get(p); 
          if (par instanceof BasicExpression) 
          { BasicExpression pbe = (BasicExpression) par; 
            if (bound.contains(pbe.data)) { } 
            else 
            { bound.add(pbe.data); } 
          } 
        } 
        
        Vector newres = new Vector(); 

        for (int j = 0; j < res.size(); j++) 
        { Expression ante = (Expression) res.get(j); 
          Expression newante = Expression.simplify("&", ante, e, null); 
          newres.add(newante); 
        } 
        res.clear(); 
        res.addAll(newres); 
      }   
      else if (conj != null && (conj instanceof BinaryExpression) && ((BinaryExpression) conj).isRuleCallDisjunction(rules))
      { Vector disjs = ((BinaryExpression) conj).allDisjuncts(); 
 
        Vector newres = new Vector(); 
                 
        for (int k = 0; k < disjs.size(); k++) 
        { BasicExpression disj = (BasicExpression) disjs.get(k); 
          Expression callk = disj.testTraceRelationWhen(entities,bound); 
          // System.out.println("*** All disjuncts = " + callk); 

          for (int j = 0; j < res.size(); j++) 
          { Expression ante = (Expression) res.get(j); 
            Expression newante = Expression.simplify("&", ante, callk, null); 
            newres.add(newante); 
          } 
        }
        res.clear(); 
        res.addAll(newres);  
      }  
      else 
      { Vector newres = new Vector(); 

        for (int j = 0; j < res.size(); j++) 
        { Expression ante = (Expression) res.get(j); 
          Expression newante = Expression.simplify("&", ante, e, null); 
          newres.add(newante); 
        } 
        res.clear(); 
        res.addAll(newres); 
      } 
    }

    // System.out.println("WHEN predicate for " + predicate + " IS " + res); 
    System.out.println(); 

    return res;
  }  // also add parameters of call to bound variables with predicate p:T if not in bound

  public Vector checkWhen(Vector bound, Vector rules, Vector entities)
  { // R(s,t) is r$trace : R$trace # s = r$trace.p1 & t = r$trace.p2
    // The s, t are bound at this point. 
    // An or of trace tests leads to multiple constraints, one for each disjunct. 

    Vector res = new Vector(); 

    Expression cond = new BasicExpression(true);
    res.add(cond); 

    for (int i = 0; i < predicate.size(); i++)
    { Expression conj = (Expression) predicate.get(i);
      Expression e = conj;
      if (conj != null && conj.isRuleCall(rules) && (conj instanceof BasicExpression))
      { BasicExpression be = (BasicExpression) conj; 
        e = be.checkWhen(entities,bound); 

        // System.out.println("RULE TEST: " + e); 

        Vector pars = be.getParameters(); 
        for (int p = 0; p < pars.size(); p++) 
        { Expression par = (Expression) pars.get(p); 
          if (par instanceof BasicExpression) 
          { BasicExpression pbe = (BasicExpression) par; 
            if (bound.contains(pbe.data)) { } 
            else 
            { bound.add(pbe.data); } 
          } 
        } 
        
        Vector newres = new Vector(); 

        for (int j = 0; j < res.size(); j++) 
        { Expression ante = (Expression) res.get(j); 
          Expression newante = Expression.simplify("&", ante, e, null); 
          newres.add(newante); 
        } 
        res.clear(); 
        res.addAll(newres); 
      }   
      else if (conj != null && (conj instanceof BinaryExpression) && ((BinaryExpression) conj).isRuleCallDisjunction(rules))
      { Vector disjs = ((BinaryExpression) conj).allDisjuncts(); 
 
        Vector newres = new Vector(); 
                 
        for (int k = 0; k < disjs.size(); k++) 
        { BasicExpression disj = (BasicExpression) disjs.get(k); 
          Expression callk = disj.checkWhen(entities,bound); 
          // System.out.println("*** All disjuncts = " + callk); 

          for (int j = 0; j < res.size(); j++) 
          { Expression ante = (Expression) res.get(j); 
            Expression newante = Expression.simplify("&", ante, callk, null); 
            newres.add(newante); 
          } 
        }
        res.clear(); 
        res.addAll(newres);  
      }  
      else 
      { Vector newres = new Vector(); 

        for (int j = 0; j < res.size(); j++) 
        { Expression ante = (Expression) res.get(j); 
          Expression newante = Expression.simplify("&", ante, e, null); 
          newres.add(newante); 
        } 
        res.clear(); 
        res.addAll(newres); 
      } 
    }

    // System.out.println("WHEN predicate for " + predicate + " IS " + res); 
    System.out.println(); 

    return res;
  }  // also add parameters of call to bound variables with predicate p:T if not in bound

  public Expression whereToExp(Vector bound, Vector targetvariables, Vector rules)
  { // invokes ops for non-top relations
    // Ignore predicates whose write frame does not involve target variables 

    Expression res = new BasicExpression(true);
    for (int i = 0; i < predicate.size(); i++)
    { Expression conj = (Expression) predicate.get(i);
      Expression e = conj;
      if (conj != null && conj.isRuleCall(rules) && 
          (conj instanceof BasicExpression))
      { e = ((BasicExpression) conj).invokeNontopRule(); }
      else if (conj instanceof BinaryExpression) 
      { BinaryExpression be = (BinaryExpression) conj; 
        /* if ("<=>".equals(be.operator))
        { // only include the direction which updates the target domains
          Vector varsleft = be.left.variablesUsedIn(targetvariables); 
          Vector varsright = be.right.variablesUsedIn(targetvariables); 
          if (varsleft.size() == 0) 
          { e = new BinaryExpression("=>", be.left, be.right);
            e.setBrackets(true); 
          } 
          else if (varsright.size() == 0) 
          { e = new BinaryExpression("=>", be.right, be.left); 
            e.setBrackets(true); 
          } 
          else 
          { System.err.println("ERROR: cannot update target variables on both sides of <=>: " + be); } 
        }
        else */  
        if ("=".equals(be.operator))
        { // Only include it if it updates a feature of a target variable
          Vector vuses = be.left.getVariableUses(); // just the updated variable. 

          System.out.println(">>>> Variable of assignment LHS: " + vuses); 
          System.out.println(">>>> Target variables: " + targetvariables); 

          e = null; 
          for (int k = 0; k < vuses.size(); k++) 
          { Expression use = (Expression) vuses.get(k); 
            if (targetvariables.contains(use + ""))
            { e = conj; 
              break; 
            } 
          }
        }  
        else 
        { e = conj; } 
      }  
          
      res = Expression.simplify("&", res, e, null);
    }
    return res;
  }
}
